'use strict';

const axios = require('axios');
const normalizeUrl = require('normalize-url');

const msPerSec = 1000;
const timeoutSeconds = 120;
const httpResponse = Object.freeze({
    continue: 100,
    ok: 200,
    badRequest: 400,
    notFound: 404,
    tooManyRequests: 429,
    unknownError: 999
});

const noRetryResponses = [httpResponse.tooManyRequests];

const normalizeLink = url => normalizeUrl(url, { defaultProtocol: 'https:', stripWWW: false, removeQueryParameters: [] });

// The ignoredLinks values are normalized when settings are processed and does not need to be repeated here.
const ignoreLink = (settings, link) => settings.ignoredLinks && settings.ignoredLinks.includes(normalizeLink(link));

const isFailedResponse = response => isNaN(response.status) || response.status >= httpResponse.badRequest;

const checkSamePageLink = async(page, link) => {
    const selector = link.slice(page.url().length);
    if (selector === '#' || selector === '#top') {
        return httpResponse.ok;
    }

    const element = await page.$(selector);
    return element ? httpResponse.ok : `${selector} Not Found`;
};

// Allow cognitive complexity less than 10
// eslint-disable-next-line sonarjs/cognitive-complexity
const checkExternalPageLink = async(page, link, useGet = false) => {
    // Get user-agent from page so axios uses the same value for requesting links
    const userAgent = await page.evaluate('navigator.userAgent');

    try {
        const options = { headers: { 'User-Agent': userAgent }, timeout: timeoutSeconds * msPerSec };
        const httpMethod = useGet ? axios.get : axios.head;
        const response = await httpMethod(link, options);
        return response.status;
    }
    catch (err) {
        // Some servers respond invalid for head request (e.g. a lot of 405 Method
        // Not Allowed), so if a response was returned, is not a response that should
        // skip a retry (e.g. 429), and the request was head then retry with get.
        if (err.response && !noRetryResponses.includes(err.response.status) && !useGet) {
            return checkExternalPageLink(page, link, true);
        }

        // Axios will throw for failed response (code >= 400), so return
        // that response code or the error code if execution error
        return err.response ? err.response.status : err.code;
    }
};

const checkLink = async(context, link) => {
    let status = httpResponse.unknownError;
    try {
        if (ignoreLink(context.testSettings, link)) {
            // Set to a unique status so obvious that it's ignored
            status = httpResponse.continue;
        }
        else if (link.startsWith(`${context.page.url()}#`)) {
            status = await checkSamePageLink(context.page, link);
        }
        else {
            status = await checkExternalPageLink(context.page, link);
        }
    }
    catch {
        status = httpResponse.unknownError;
    }
    return status;
};

module.exports.checkLink = checkLink;
module.exports.httpResponse = httpResponse;
module.exports.isFailedResponse = isFailedResponse;
module.exports.normalizeLink = normalizeLink;
