'use strict';

const axios = require('axios');
const puppeteer = require('puppeteer');
const protocolify = require('protocolify');

const { checkLink, httpResponse, isFailedResponse, normalizeLink } = require('../lib/linkUtils');

// eslint-disable-next-line max-lines-per-function
describe('check link', () => {
    const testUrl = './tests/test-cases/brokenLinks.html';
    const testUrlLink = protocolify(testUrl);
    const userAgent = 'pagean test user-agent';
    const puppeteerLaunchOptions = { args: ['--no-sandbox'] };
    let browser, page;

    const getContext = (ignoredLinks = []) => {
        return {
            testSettings: {
                ignoredLinks
            },
            page
        };
    };

    beforeAll(async() => {
        browser = await puppeteer.launch(puppeteerLaunchOptions);
        page = await browser.newPage();
        page.setUserAgent(userAgent);
        await page.goto(testUrlLink, { waitUntil: 'load' });
    });

    afterEach(() => {
        jest.restoreAllMocks();
    });

    afterAll(async() => {
        await page.close();
        await browser.close();
    });

    // eslint-disable-next-line max-lines-per-function
    const checkLinksWithMocks = async(link, ignoredLinks, invalidPage = false, httpStatus = httpResponse.ok) => {
        const context = getContext(ignoredLinks);
        if (invalidPage) {
            // This enables forcing an error to be thrown for various test cases
            delete context.page;
        }
        const pageGetElementSpy = jest.spyOn(page, '$');
        const pageEvaluateSpy = jest.spyOn(page, 'evaluate');

        const axiosMock = () => {
            if (isNaN(httpStatus)) {
                // If execution error, format to match axios error
                const err = new Error();
                err.code = httpStatus;
                throw err;
            }
            else if (httpStatus >= httpResponse.badRequest) {
                // If erroneous response, throw error with response to match axios
                const err = new Error();
                err.response = { status: httpStatus };
                throw err;
            }
            return { status: httpStatus };
        };
        const axiosHeadSpy = jest.spyOn(axios, 'head').mockImplementation(axiosMock);
        const axiosGetSpy = jest.spyOn(axios, 'get').mockImplementation(axiosMock);

        const result = await checkLink(context, link);

        return {
            axiosGetSpy,
            axiosHeadSpy,
            pageEvaluateSpy,
            pageGetElementSpy,
            result
        };
    };

    // eslint-disable-next-line max-lines-per-function
    describe('ignored links', () => {
        const ignoreLinksTestLink = 'https://this.is.a.test/';
        const normalizedIgnoreLinksTestLink = 'https://this.is.a.test';

        it('should return 100 if exact link is in ignored links', async() => {
            expect.assertions(1);
            const ignoredLinks = [normalizedIgnoreLinksTestLink];
            const testResults = await checkLinksWithMocks(ignoreLinksTestLink, ignoredLinks);
            expect(testResults.result).toStrictEqual(httpResponse.continue);
        });

        it('should return 100 if normalized link is in ignored links', async() => {
            expect.assertions(1);
            const ignoredLinks = [normalizedIgnoreLinksTestLink];
            const testResults = await checkLinksWithMocks(normalizedIgnoreLinksTestLink, ignoredLinks);
            expect(testResults.result).toStrictEqual(httpResponse.continue);
        });

        it('should not return 100 is link is not in ignored links', async() => {
            expect.assertions(1);
            const ignoredLinks = ['https://this.is.not.a.test/'];
            const testResults = await checkLinksWithMocks(ignoreLinksTestLink, ignoredLinks);
            expect(testResults.result).not.toStrictEqual(httpResponse.continue);
        });

        it('should not return 100 is ignored links are not specified', async() => {
            expect.assertions(1);
            const testResults = await checkLinksWithMocks(ignoreLinksTestLink);
            expect(testResults.result).not.toStrictEqual(httpResponse.continue);
        });
    });

    // eslint-disable-next-line max-lines-per-function
    describe('same page links', () => {
        it('should return 200 for url with # and no element (top of page)', async() => {
            expect.assertions(1);
            const link = `${testUrlLink}#`;
            const testResults = await checkLinksWithMocks(link);
            expect(testResults.result).toStrictEqual(httpResponse.ok);
        });

        it('should return 200 for url with #top', async() => {
            expect.assertions(1);
            const link = `${testUrlLink}#top`;
            const testResults = await checkLinksWithMocks(link);
            expect(testResults.result).toStrictEqual(httpResponse.ok);
        });

        it('should check page for element if element specified', async() => {
            expect.assertions(1);
            const element = '#foo';
            const link = `${testUrlLink}${element}`;
            const testResults = await checkLinksWithMocks(link);
            expect(testResults.pageGetElementSpy).toHaveBeenCalledWith(element);
        });

        it('should return 200 if element found on page', async() => {
            expect.assertions(1);
            const element = '#linked';
            const link = `${testUrlLink}${element}`;
            const testResults = await checkLinksWithMocks(link);
            expect(testResults.result).toStrictEqual(httpResponse.ok);
        });

        it('should return #element not found error if element is not found on page', async() => {
            expect.assertions(1);
            const element = '#notlinked';
            const link = `${testUrlLink}${element}`;
            const testResults = await checkLinksWithMocks(link);
            expect(testResults.result).toStrictEqual(`${element} Not Found`);
        });

        it('should return 999 if error thrown while checking link', async() => {
            expect.assertions(1);
            const element = '#linked';
            const link = `${testUrlLink}${element}`;
            const testResults = await checkLinksWithMocks(link, [], true);
            expect(testResults.result).toStrictEqual(httpResponse.unknownError);
        });
    });

    // eslint-disable-next-line max-lines-per-function
    describe('external page links', () => {
        const externalLink = 'https://www.google.com';

        it('should make only a HEAD request for link if not same page and response is valid', async() => {
            expect.assertions(2); // eslint-disable-line no-magic-numbers
            const testResults = await checkLinksWithMocks(externalLink);
            expect(testResults.axiosHeadSpy.mock.calls[0][0]).toStrictEqual(externalLink);
            expect(testResults.axiosGetSpy).not.toHaveBeenCalled();
        });

        it('should make only a HEAD request for link if not same page and HEAD returns code 429', async() => {
            expect.assertions(2); // eslint-disable-line no-magic-numbers
            const testResults = await checkLinksWithMocks(externalLink, [], false, httpResponse.tooManyRequests);
            expect(testResults.axiosHeadSpy.mock.calls[0][0]).toStrictEqual(externalLink);
            expect(testResults.axiosGetSpy).not.toHaveBeenCalled();
        });

        it('should make only a HEAD request for link if not same page and HEAD returns an execution error', async() => {
            expect.assertions(2); // eslint-disable-line no-magic-numbers
            const responseCode = 'ENOTFOUND';
            const testResults = await checkLinksWithMocks(externalLink, [], false, responseCode);
            expect(testResults.axiosHeadSpy.mock.calls[0][0]).toStrictEqual(externalLink);
            expect(testResults.axiosGetSpy).not.toHaveBeenCalled();
        });

        it('should make only a HEAD request passing the current page user agent if response is valid', async() => {
            expect.assertions(2); // eslint-disable-line no-magic-numbers
            const expectedHeaders = { headers: { 'User-Agent': userAgent } };
            const testResults = await checkLinksWithMocks(externalLink);
            expect(testResults.axiosHeadSpy.mock.calls[0][1]).toStrictEqual(expect.objectContaining(expectedHeaders));
            expect(testResults.axiosGetSpy).not.toHaveBeenCalled();
        });

        it('should make HEAD and GET request for link if not same page and HEAD returns error', async() => {
            expect.assertions(2); // eslint-disable-line no-magic-numbers
            const testResults = await checkLinksWithMocks(externalLink, [], false, httpResponse.badRequest);
            expect(testResults.axiosHeadSpy.mock.calls[0][0]).toStrictEqual(externalLink);
            expect(testResults.axiosGetSpy.mock.calls[0][0]).toStrictEqual(externalLink);
        });

        it('should make HEAD and GET request passing the current page user agent if HEAD returns error', async() => {
            expect.assertions(2); // eslint-disable-line no-magic-numbers
            const expectedHeaders = { headers: { 'User-Agent': userAgent } };
            const testResults = await checkLinksWithMocks(externalLink, [], false, httpResponse.badRequest);
            expect(testResults.axiosHeadSpy.mock.calls[0][1]).toStrictEqual(expect.objectContaining(expectedHeaders));
            expect(testResults.axiosGetSpy.mock.calls[0][1]).toStrictEqual(expect.objectContaining(expectedHeaders));
        });

        it('should return the HTTP response status for a nominal request', async() => {
            expect.assertions(1);
            const testResults = await checkLinksWithMocks(externalLink, [], false, httpResponse.ok);
            expect(testResults.result).toStrictEqual(httpResponse.ok);
        });

        it('should return the HTTP response status for an off-nominal request with a response', async() => {
            expect.assertions(1);
            const link = 'https://www.google.com/this.is.not.found/';
            const testResults = await checkLinksWithMocks(link, [], false, httpResponse.notFound);
            expect(testResults.result).toStrictEqual(httpResponse.notFound);
        });

        it('should return the error code for a request with an execution error', async() => {
            expect.assertions(1);
            const link = 'https://this.is.a.test/';
            const responseCode = 'ENOTFOUND';
            const testResults = await checkLinksWithMocks(link, [], false, responseCode);
            expect(testResults.result).toStrictEqual(responseCode);
        });
    });
});

describe('is failed response', () => {
    const response = { status: undefined };
    it('should return true for any numeric status value greater than or equal to 400', () => {
        // eslint-disable-next-line no-magic-numbers
        expect.assertions(2);
        response.status = 400;
        expect(isFailedResponse(response)).toStrictEqual(true);
        response.status = 500;
        expect(isFailedResponse(response)).toStrictEqual(true);
    });

    it('return false for any numeric status value less than 400', () => {
        // eslint-disable-next-line no-magic-numbers
        expect.assertions(2);
        response.status = 302;
        expect(isFailedResponse(response)).toStrictEqual(false);
        response.status = 200;
        expect(isFailedResponse(response)).toStrictEqual(false);
    });

    it('return true for any status value that is nor a number', () => {
        // eslint-disable-next-line no-magic-numbers
        expect.assertions(2);
        response.status = 'ENOTFOUND';
        expect(isFailedResponse(response)).toStrictEqual(true);
        response.status = undefined;
        expect(isFailedResponse(response)).toStrictEqual(true);
    });
});

describe('normalize link', () => {
    const referenceUrl = 'https://www.test.com';

    it('should remove trailing slash without altering URL', () => {
        expect.assertions(1);
        const initialURL = `${referenceUrl}/`;
        expect(normalizeLink(initialURL)).toStrictEqual(referenceUrl);
    });

    it('should not strip www from domain', () => {
        expect.assertions(1);
        expect(normalizeLink(referenceUrl)).toStrictEqual(referenceUrl);
    });

    it('should add https as protocol if protocol relative URL', () => {
        expect.assertions(1);
        const initialURL = referenceUrl.replace('https:', '');
        expect(normalizeLink(initialURL)).toStrictEqual(referenceUrl);
    });

    it('should not remove hash from url', () => {
        expect.assertions(1);
        const initialURL = `${referenceUrl}/#elementId`;
        expect(normalizeLink(initialURL)).toStrictEqual(initialURL);
    });
});
